#include "include/common.h"

static char	*extend_sql[] = {
	"DUMP TO ",
	"SHOW ",
	"DESC "
};

static int	extend_sql_len[] = {
	8, 5, 5
};

static int	extend_sql_cnt = 3;

int extend_process( struct thread_t *self, int db_open_res )
{
	extern struct config_t	config;

	extern regex_t		reg_admin_dump;
	extern regex_t		reg_admin_show;
	extern regex_t		reg_admin_desc;

	extern char		*libversion;

	char			*query;
	int			i;

	regmatch_t		match[5];
	int			match_res;

	char			*value;

	char			*sql;
	int			process_res;


	if ( self->request.data_len <= 0)
		return SQLITE_ERROR_ADMIN;

	query = self->request.data;
	query += strspn( query, " \r\n\t" );

	if ( ( query[0] == 's' || query[0] == 'S' ) && query[1] != 'h' && query[1] != 'H' )
		return SQLITE_ERROR_EXTEND;

	if ( query[0] != 's' && query[0] != 'S' && query[0] != 'd' && query[0] != 'D' )
		return SQLITE_ERROR_EXTEND;

#ifdef __DEBUG
	printf("***** Extended SQL: %s\n", query);
#endif

	for ( i=extend_sql_cnt; i--; )
	{
		if ( strncasecmp( extend_sql[i], query, extend_sql_len[i] ) == 0 )
		{
			switch (i)
			{
				case 0:
					//DUMP TO [[MIRROR] FILE 'filename' | CLIENT]
					match_res = regexec( &reg_admin_dump, query, 6, match, 0 );
					if ( match_res == 0 )
					{
						if ( db_open_res == SQLITE_NOTFOUND )
							return response_error( self, RESPONSE_ERR_DB_FOUND, NULL );
					}
					//send error and return
					return response_error( self, RESPONSE_ERR_INTERNAL, "\"DUMP\" SQL is not support yet." );
				break;
				case 1:
					//SHOW
					if ( config.access_control == TRUE && authorizer( &self->user, SQLITE_SHOW, NULL, NULL, NULL, NULL ) != SQLITE_OK )
						return response_error( self, RESPONSE_ERR_INTERNAL, "not authorized" );

					match_res = regexec( &reg_admin_show, query, 3, match, 0 );
					if ( match_res == 0 )
					{
						if ( strncasecmp( &(query[match[1].rm_so]), "TABLE", 5 ) == 0 )
						{
							if ( db_open_res == SQLITE_NOTFOUND )
								return response_error( self, RESPONSE_ERR_DB_FOUND, NULL );
							/*
							SHOW TABLES [table_name]
							SELECT type, name, sql FROM sqlite_master WHERE type IN ('table','view') and name like '%%%q%%' UNION ALL SELECT name FROM sqlite_temp_master WHERE type IN ('table','view') and name like '%%%q%%' ORDER BY name
							*/
							if ( match[2].rm_so > 0 && match[2].rm_eo > 0 && match[2].rm_so < match[2].rm_eo )
							{
								value = &(query[match[2].rm_so]);
								query[match[2].rm_eo] = '\0';
								sql = db_printf( "SELECT type, name, sql FROM sqlite_master WHERE type IN ('table','view') and name='%q' UNION ALL SELECT type, name, sql FROM sqlite_temp_master WHERE type IN ('table','view') and name='%q' ORDER BY name", value, value );
							}
							else
								sql = db_printf( "SELECT type, name, sql FROM sqlite_master WHERE type IN ('table','view') UNION ALL SELECT type, name, sql FROM sqlite_temp_master WHERE type IN ('table','view') ORDER BY name" );
							process_res = client_process( self, sql );
							db_free( sql );
							return process_res;
						}
						else if ( strncasecmp( &(query[match[1].rm_so]), "INDEX", 5 ) == 0 )
						{
							if ( db_open_res == SQLITE_NOTFOUND )
								return response_error( self, RESPONSE_ERR_DB_FOUND, NULL );
							/*
							SHOW INDEX [table_name]
							SELECT type, name, sql FROM sqlite_master WHERE type='index' AND tbl_name='%q' UNION ALL SELECT type, name, sql FROM sqlite_temp_master WHERE type='index' AND tbl_name='%q' ORDER BY name
							*/
							if ( match[2].rm_so > 0 && match[2].rm_eo > 0 && match[2].rm_so < match[2].rm_eo )
							{
								value = &(query[match[2].rm_so]);
								query[match[2].rm_eo] = '\0';
								sql = db_printf( "SELECT type, name, tbl_name, sql FROM sqlite_master WHERE type='index' AND tbl_name='%q' UNION ALL SELECT type, name, tbl_name, sql FROM sqlite_temp_master WHERE type='index' AND tbl_name='%q' ORDER BY name", value, value );
							}
							else
								sql = db_printf( "SELECT type, name, tbl_name, sql FROM sqlite_master WHERE type='index' UNION ALL SELECT type, name, tbl_name, sql FROM sqlite_temp_master WHERE type='index' ORDER BY name" );
							process_res = client_process( self, sql );
							db_free( sql );
							return process_res;
						}
						else if ( strncasecmp( &(query[match[1].rm_so]), "STATUS", 6 ) == 0 )
							return monitor_response( self );
						else if ( strncasecmp( &(query[match[1].rm_so]), "TRIGGER", 7 ) == 0 )
						{
							//SHOW TRIGGER [table_name]
							if ( match[2].rm_so > 0 && match[2].rm_eo > 0 && match[2].rm_so < match[2].rm_eo )
							{
								value = &(query[match[2].rm_so]);
								query[match[2].rm_eo] = '\0';
								sql = db_printf( "SELECT type, name, tbl_name, sql FROM sqlite_master WHERE type='trigger' AND tbl_name='%q' UNION ALL SELECT type, name, tbl_name, sql FROM sqlite_temp_master WHERE type='trigger' AND tbl_name='%q' ORDER BY name", value, value );
							}
							else
								sql = db_printf( "SELECT type, name, tbl_name, sql FROM sqlite_master WHERE type='trigger' UNION ALL SELECT type, name, tbl_name, sql FROM sqlite_temp_master WHERE type='trigger' ORDER BY name" );
							process_res = client_process( self, sql );
							db_free( sql );
							return process_res;
						}
						else if ( strncasecmp( &(query[match[1].rm_so]), "VERSION", 7 ) == 0 )
							return response_message( self, libversion );
						else if ( strncasecmp( &(query[match[1].rm_so]), "DATABASE", 8 ) == 0 )
							return extend_show_database( self );
						else if ( strncasecmp( &(query[match[1].rm_so]), "USER", 4 ) == 0 )
							return extend_show_user( self );
					}
					//send error and return OK
					return response_error( self, RESPONSE_ERR_INTERNAL, "\"SHOW\" SQL syntax not valid." );
				break;
				case 2:
					//DESC
					if ( config.access_control == TRUE && authorizer( &self->user, SQLITE_SHOW, NULL, NULL, NULL, NULL ) != SQLITE_OK )
						return response_error( self, RESPONSE_ERR_INTERNAL, "not authorized" );

					match_res = regexec( &reg_admin_desc, query, 2, match, 0 );
					if ( match_res == 0 )
					{
						if ( db_open_res == SQLITE_NOTFOUND )
							return response_error( self, RESPONSE_ERR_DB_FOUND, NULL );

						value = &(query[match[1].rm_so]);
						query[match[1].rm_eo] = '\0';

						return extend_desc( self, value );
					}
					return response_error( self, RESPONSE_ERR_INTERNAL, "\"DESC\" SQL syntax not valid." );
				break;
			}
		}
	}

	return SQLITE_ERROR_EXTEND;
}

#define R_BUF		&(self->response.buf[self->response.offset])
#define R_BUF_LIMIT	(self->response.buf_size-self->response.offset)
int extend_desc( struct thread_t *self, char *tablename )
{
	int		result;
	char		*sql;
	int		change;
	long long int	rowid;
	int		column_count;
	int		i;
	const char	*value;
	struct desc_t	desc;

	sql = db_printf( "SELECT * FROM %s LIMIT 0", tablename );

	change = db_changes( &self->db );
	rowid = db_last_rowid( &self->db );

	result = db_statement( &self->db, sql );
	db_free( sql );
	if ( result != SQLITE_OK )
		return response_error( self, RESPONSE_ERR_INTERNAL, (char *) db_error( &self->db ) );

	column_count = db_statement_column_count( &self->db );

	self->response.start = self->response.offset = 20;

	self->response.offset += response_buf_open( R_BUF, R_BUF_LIMIT, change, rowid, column_count, &self->bench_start );

	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_OPEN );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "name" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "type" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
#ifdef SQLITE_ENABLE_COLUMN_METADATA
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "seq" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "notnull" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "pkey" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "autoincrement" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
#endif

	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_CLOSE );

	for ( i=0; i < column_count; i++ )
	{
		value = db_statement_column_name( &self->db, i );
		if ( db_column_meta( &self->db, tablename, value, &desc ) == SQLITE_OK )
		{
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_OPEN );
				//column name
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
				self->response.offset += response_xml_printf( R_BUF, value, strlen(value) );
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );

				//column type
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
				if ( desc.type != NULL )
					self->response.offset += response_xml_printf( R_BUF, desc.type, strlen(desc.type) );
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );

				//column seq
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
				if ( desc.seq != NULL )
					self->response.offset += response_xml_printf( R_BUF, desc.seq, strlen(desc.seq) );
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );

				//column not null
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
				if ( desc.notnull == 1 )
					self->response.offset += response_xml_printf( R_BUF, "O", 1 );
				else
					self->response.offset += response_xml_printf( R_BUF, "X", 1 );
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );

				//column primary key
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
				if ( desc.pkey == 1 )
					self->response.offset += response_xml_printf( R_BUF, "O", 1 );
				else
					self->response.offset += response_xml_printf( R_BUF, "X", 1 );
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );

				//column auto-increment
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
				if ( desc.autoincrement == 1 )
					self->response.offset += response_xml_printf( R_BUF, "O", 1 );
				else
					self->response.offset += response_xml_printf( R_BUF, "X", 1 );
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );

			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_CLOSE );
		}
		else
		{
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_OPEN );
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
					self->response.offset += response_xml_printf( R_BUF, value, strlen(value) );
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
					value = db_statement_column_type( &self->db, i );
					if ( value != NULL )
						self->response.offset += response_xml_printf( R_BUF, value, strlen(value) );
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_CLOSE );
		}
	}
	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_CLOSE );

	self->response.status = 200;
	self->response.status_msg = "OK";

	db_statement_free( &self->db );

	return response( self );
}

int extend_show_database( struct thread_t *self )
{
	extern struct config_t	config;

	long long int		rowid;

	struct dirent		**dirlist;
	int			result;
	int			database_cnt;
	int			i;

	char			filepath[1024];
	struct stat		filestat;

	rowid = db_last_rowid( &self->db );

	result = scandir( config.db_path, &dirlist, 0, alphasort );
	if ( result < 0 )
		return response_error( self, RESPONSE_ERR_INTERNAL, "Cannot open data directory." );

	i = result;
	database_cnt = 0;
	while ( i-- )
	{
		if ( strcmp( dirlist[i]->d_name, "catalog.user.superdb" ) == 0 )
		{
			(dirlist[i]->d_name[0]) = 0;
			continue;
		}

		snprintf( filepath, 1024, "%s/%s", config.db_path, dirlist[i]->d_name );
		if ( lstat( filepath, &filestat ) == 0 )
		{
			if ( S_ISREG( filestat.st_mode ) || S_ISLNK( filestat.st_mode ) )
				database_cnt++;
			else
				(dirlist[i]->d_name[0]) = 0;
		}
		else
			(dirlist[i]->d_name[0]) = 0;
	}

	self->response.start = self->response.offset = 20;
	self->response.offset += response_buf_open( R_BUF, R_BUF_LIMIT, 0, rowid, database_cnt, &self->bench_start );

	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_OPEN );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "name" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_CLOSE );

	// :memory:
	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_OPEN );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
			self->response.offset += response_xml_printf( R_BUF, DB_MEMORY, DB_MEMORY_LEN );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );
	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_CLOSE );

	i = result;
	while ( i-- )
	{
		if ( (dirlist[i]->d_name[0]) != 0 )
		{
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_OPEN );
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
					self->response.offset += response_xml_printf( R_BUF, dirlist[i]->d_name, strlen(dirlist[i]->d_name) );
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_CLOSE );
		}
	}
	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_CLOSE );

	self->response.status = 200;
	self->response.status_msg = "OK";

	while ( result-- )
		free( dirlist[result] );

	free( dirlist );

	return response( self );
}

#define	AUTH_FIELD_ID		0
#define	AUTH_FIELD_SUPER	2
int extend_show_user( struct thread_t *self )
{
	extern struct auth_t	auth;
	extern pthread_mutex_t	db_auth_mutex;

	long long int		rowid;
	int			i;

	rowid = db_last_rowid( &self->db );

	self->response.start = self->response.offset = 20;
	self->response.offset += response_buf_open( R_BUF, R_BUF_LIMIT, 0, rowid, auth.user.rows, &self->bench_start );

	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_OPEN );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "name" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "superuser" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_CLOSE );

	pthread_mutex_lock( &db_auth_mutex );
	for ( i=0; i < auth.user.rows; i++ )
	{
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
				self->response.offset += response_xml_printf( R_BUF, 
					auth.user.table[(auth.user.cols * (i+1) + AUTH_FIELD_ID)], 
					strlen(auth.user.table[(auth.user.cols * (i+1) + AUTH_FIELD_ID)]) );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
				self->response.offset += response_xml_printf( R_BUF, 
					auth.user.table[(auth.user.cols * (i+1) + AUTH_FIELD_SUPER)], 1 );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_CLOSE );
	}
	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_CLOSE );
	pthread_mutex_unlock( &db_auth_mutex );

	self->response.status = 200;
	self->response.status_msg = "OK";

	return response( self );
}
#undef	AUTH_FIELD_ID
#undef	AUTH_FIELD_SUPER

#undef R_BUF
#undef R_BUF_LIMIT
