#include "include/common.h"

#define	CLIENT_KEEP_OR_BREAK	{if ( self->keep_alive == TRUE ) {	\
					client_session_init( self );	\
					goto request_continue;		\
				}else break;}

#define	RESPONSEERROR(x)	if ( response_error( self, x, NULL ) != SQLITE_OK ) {break;}

void *client( void *data )
{
	extern struct config_t	config;
	extern struct func_t	func;
	extern long long	thread_seq;
	extern pthread_mutex_t	thread_seq_mutex;
	extern short		terminate;
	extern int		errno;

	struct thread_t		*self = ((struct thread_t *) data);

	struct pollfd		pollfd;
	int			poll_res;
	int			read_res;

	int			auth_res;

	int			db_res;
	int			result;

	int			http_res;

#ifdef	ENABLE_SSL
	extern SSL_CTX		*ssl_ctx;
#endif

	self->init = FALSE;
	self->status = THREAD_INIT;

	time( &(self->time_init) );
	self->time_last = 0;

	if ( self->client_fd <= 0 )
	{
		if ( func.on_thread_exit != NULL )
			func.on_thread_exit( self );
		free( self );
		pthread_exit(0);
	}

	pollfd.fd = self->client_fd;

#ifdef	ENABLE_SSL
	if ( config.ssl == TRUE )
	{
		self->ssl = SSL_new( ssl_ctx );
		if ( self->ssl == NULL )
		{
			close( self->client_fd );
			if ( func.on_thread_exit != NULL )
				func.on_thread_exit( self );
			free( self );
			pthread_exit( 0 );
		}

		if ( SSL_set_fd( self->ssl, self->client_fd ) == 0 )
		{
			SSL_free( self->ssl );
			close( self->client_fd );
			if ( func.on_thread_exit != NULL )
				func.on_thread_exit( self );
			free( self );
			pthread_exit( 0 );
		}

		result = SSL_accept( self->ssl );
		if ( result == -1 )
		{
			while ( ( result = SSL_do_handshake( self->ssl ) ) != 1 )
			{
				if ( result == 0 )
					break;
				if ( SSL_get_error( self->ssl, result ) == SSL_ERROR_WANT_READ )
					pollfd.events = POLLIN;
				else if ( SSL_get_error( self->ssl, result ) == SSL_ERROR_WANT_READ )
					pollfd.events = POLLOUT;
				else
					break;
				result = poll( &pollfd, 1, 1000 );
			}
		}

		if ( result != 1 )
		{
			SSL_free( self->ssl );
			close( self->client_fd );
			if ( func.on_thread_exit != NULL )
				func.on_thread_exit( self );
			free( self );
			pthread_exit( 0 );
		}
	}
#endif
	pollfd.events = POLLIN | POLLPRI | POLLERR | POLLHUP;

	self->request.buf_size = config.buf_size_recv;
	self->request.buf = malloc( self->request.buf_size );

	self->response.header_buf = malloc( config.buf_size_header );

	self->response.buf_size = config.buf_size_send;
	self->response.buf = malloc( self->response.buf_size );

	pthread_mutex_lock( &thread_seq_mutex );
	self->tid = ++thread_seq;
	pthread_mutex_unlock( &thread_seq_mutex );

	monitor_thread_push( self );
	self->init = TRUE;

	/* execute external function */
	if ( func.after_thread != NULL )
	{
		result = func.after_thread( self );
		if ( result == STAT_PROC_ERROR || result == STAT_PROC_GO_EXIT )
			goto client_error;
	}

	while ( 1 )
	{
		/* execute external function */
		if ( func.after_proc != NULL )
		{
			result = func.after_proc( self );
			if ( result == STAT_PROC_ERROR || result == STAT_PROC_GO_EXIT )
				break;
		}

request_continue:
		if ( terminate == TRUE )
			break;
		self->status = THREAD_WAIT;
		poll_res = poll( &pollfd, 1, config.http_keeptime*1000 );

		if ( config.http_timeout <= 0 && poll_res == 0 )
			continue;
		if ( poll_res < 0 )
		{
			log_server_print( "*** Polling failed: %s", strerror(errno) );
			break;
		}

		if ( self->request.buf_size-self->request.read_len <= 1024*8 )
		{
			self->request.buf = realloc( self->request.buf, self->request.buf_size + LIMIT_BUF_INCREMENT );
			self->request.buf_size += LIMIT_BUF_INCREMENT;
		}

#ifdef	ENABLE_SSL
		if ( config.ssl == TRUE )
			read_res = SSL_read( self->ssl, &(self->request.buf[self->request.read_len]), self->request.buf_size-self->request.read_len );
		else
#endif
			read_res = read( self->client_fd, &(self->request.buf[self->request.read_len]), self->request.buf_size-self->request.read_len );

		if ( read_res <= 0 )
			break;

		self->request.read_len += read_res;

		if ( self->request.header_len <= 0 )
		{
			self->request.buf[self->request.read_len] = '\0';
#ifdef __DEBUG
			//printf("%s\n", self->request.buf);
#endif
			if ( request_header_parse( &(self->request) ) == SQLITE_ERROR )
			{
				response_error( self, RESPONSE_ERR_BAD_REQUEST, NULL );
				goto client_error;
			}
		}

		if ( self->request.header_len > 0 ) 
		{
			switch ( request_end( &(self->request) ) )
			{
				case SQLITE_OK:
				break;
				case SQLITE_EMPTY:
					goto request_continue;
				break;
				case SQLITE_ERROR:
					if ( self->request.header.method == REQUEST_METHOD_GET )
						break;
					response_error( self, RESPONSE_ERR_BAD_REQUEST, NULL );
					goto client_error;
				break;
			}

			self->request.buf[self->request.read_len] = '\0';
			self->keep_alive = request_keep_alive( &(self->request) );
			self->status = THREAD_WORK;

			/* execute external function */
			if ( func.after_read != NULL )
			{
				result = func.after_read( self );
				if ( result == STAT_PROC_ERROR || result == STAT_PROC_GO_EXIT )
					break;
				else if ( result == STAT_PROC_GO_READ )
					goto request_continue;
			}

			/* POST, OPTION */
			if ( self->request.header.method == REQUEST_METHOD_POST || self->request.header.method == REQUEST_METHOD_OPTION )
			{
				/* execute external function */
				if ( func.on_post != NULL )
				{
					result = func.on_post( self );
					if ( result == STAT_PROC_ERROR || result == STAT_PROC_GO_EXIT )
						break;
					else if ( result == STAT_PROC_GO_READ )
						goto request_continue;
				}

				time( &(self->time_last) );
				gettimeofday( &(self->bench_start), NULL );

				auth_res = client_auth_proc( self );

				if ( auth_res != SQLITE_OK )
				{
					//auth error
					RESPONSEERROR(RESPONSE_ERR_UNAUTHORIZED)
					CLIENT_KEEP_OR_BREAK
				}

				db_res = client_open_db( self );

				if ( self->request.header.method == REQUEST_METHOD_POST )
				{
					if ( db_res == SQLITE_NOTFOUND || db_res == SQLITE_OK )
					{
						result = extend_process( self, db_res );
						if ( result == SQLITE_OK )
							CLIENT_KEEP_OR_BREAK
						else if ( result == SQLITE_ERROR_SEND )
							break;

						if ( self->user.super == 1 )
						{
							result = admin_process( self, db_res );
							if ( result == SQLITE_OK )
								CLIENT_KEEP_OR_BREAK
							else if ( result == SQLITE_ERROR_SEND )
								break;
						}
					}

					if ( db_res == SQLITE_OK )
					{
						//db process
						if ( self->request.data_len <= 0 )
						{
							if ( response_info( self, RESPONSE_INFO_LOGIN_SUCCESS, NULL ) != SQLITE_OK )
								break;
							CLIENT_KEEP_OR_BREAK
						}

						result = client_process( self, self->request.data );
						if ( result == SQLITE_OK )
							CLIENT_KEEP_OR_BREAK
						else if ( result == SQLITE_ERROR_SEND )
							break;
					}
					else
					{
						RESPONSEERROR(RESPONSE_ERR_DB_FOUND)
						CLIENT_KEEP_OR_BREAK
					}

				}
				else
				/* prepared statement */
				{
					if ( db_res != SQLITE_OK )
					{
						RESPONSEERROR(RESPONSE_ERR_DB_FOUND)
						CLIENT_KEEP_OR_BREAK
					}

					result = pstmt_process( self );
					if ( result == SQLITE_OK )
						CLIENT_KEEP_OR_BREAK
					else if ( result == SQLITE_ERROR_SEND )
						break;
				}
			}
			/* PUT, DELETE for WebDAV */
			else if ( self->request.header.method == REQUEST_METHOD_PUT || self->request.header.method == REQUEST_METHOD_DELETE )
			{
				/* execute external function */
				if ( func.on_webdav != NULL )
				{
					result = func.on_webdav( self );
					if ( result == STAT_PROC_ERROR || result == STAT_PROC_GO_EXIT )
						break;
					else if ( result == STAT_PROC_GO_READ )
						goto request_continue;
				}
					
				auth_res = client_auth_proc( self );
				if ( auth_res != SQLITE_OK || self->user.super != 1 )
				{
					//auth error
					RESPONSEERROR(RESPONSE_ERR_UNAUTHORIZED)
					CLIENT_KEEP_OR_BREAK
				}

				if ( config.webdav == TRUE )
				{
					http_res = webdav_proc( self );
					if ( http_res == SQLITE_OK )
					{
						CLIENT_KEEP_OR_BREAK
					}
					else
						break;
				}
				else
				{
					RESPONSEERROR(RESPONSE_ERR_BAD_REQUEST)
					CLIENT_KEEP_OR_BREAK
				}
			}
			/* GET or etc.. */
			else
			{
				/* execute external function */
				if ( func.on_get != NULL )
				{
					result = func.on_get( self );
					if ( result == STAT_PROC_ERROR || result == STAT_PROC_GO_EXIT )
						break;
					else if ( result == STAT_PROC_GO_READ )
						goto request_continue;
				}

				if ( config.http_use == TRUE )
				{
					http_res = http_proc( self );
					if ( http_res == SQLITE_OK )
					{
						CLIENT_KEEP_OR_BREAK
					}
					else
						break;
				}
				else
				{
					CLIENT_KEEP_OR_BREAK
				}
			}
			CLIENT_KEEP_OR_BREAK
		}
	}

client_error:

	log_server( "**** Client thread terminated." );
#ifdef __DEBUG
	printf("***** Client thread terminated.\n");
#endif

	if ( func.on_thread_exit != NULL )
		func.on_thread_exit( self );

	client_safe_exit( self );

	db_thread_cleanup();

	pthread_exit( 0 );
}

int client_auth_proc( struct thread_t *self )
{
	char				*cookie_value;
	short				cookie_auth;
	char				*base64;

	char				id[1024];
	char				*pass;

	struct timeval			now;
	char				tmp[64];
	int				tmp_len;

	cookie_auth = FALSE;

	if ( self->request.header.cookie != NULL && strlen(self->session_key) > 0 )
	{
		cookie_value = strstr( self->request.header.cookie, "SQLSESSION=" );
		if ( cookie_value != NULL )
		{
			cookie_value += 11;
			if ( strncmp( cookie_value, self->session_key, MINVAL( 32, MINVAL( strlen(cookie_value), strlen(self->session_key) ) ) ) == 0 )
				return SQLITE_OK;
		}
	}

	if ( self->request.header.auth == NULL )
		return SQLITE_AUTH;

	base64 = strchr( self->request.header.auth, ' ' );
	if ( base64 == NULL )
		return SQLITE_AUTH;

	base64++;

	if ( auth_parse_str( id, &pass, base64 ) != SQLITE_OK )
		return SQLITE_AUTH;

	if ( auth_user_check( id, pass, &self->user ) == SQLITE_OK )
	{
		gettimeofday( &now, NULL );
#if defined(__APPLE__)
		tmp_len = snprintf( tmp, 64, "%uld%ld", now.tv_usec, now.tv_sec );
#else
		tmp_len = snprintf( tmp, 64, "%ld%ld", now.tv_usec, now.tv_sec );
#endif
		for ( ; tmp_len--; )
			tmp[tmp_len] += ( 17 + tmp[tmp_len]%(tmp_len+1) );

		snprintf( self->session_key, 32, "%s", tmp );
		self->response.set_cookie = 1;
		return SQLITE_OK;
	}
	else
	{
		return SQLITE_AUTH;
	}
}

int client_open_db( struct thread_t *self )
{
	extern struct config_t	config;
	extern short		repl_master_able;
	extern int		errno;
	int			open_res;

	if ( strlen( self->request.header.db_name ) <= 0 )
	{
		log_server_print( "*** Database not found: %s", self->request.header.db_name );
		return SQLITE_NOTFOUND;
	}

	self->request.header.extend = strchr( self->request.header.db_name, '/' );
	if ( self->request.header.extend != NULL )
	{
		self->request.header.extend[0] = '\0';
		self->request.header.extend++;
		self->request.header.db_name_len = strlen( self->request.header.db_name );
	}

	if ( strlen(self->db.db_name) == self->request.header.db_name_len
		&& strncasecmp( self->db.db_name, self->request.header.db_name, self->request.header.db_name_len ) == 0 )
	{
		return SQLITE_OK;
	}

	db_close( &self->db );
	open_res = db_open( &self->db, self->request.header.db_name );

	if ( open_res == SQLITE_OK )
	{
		memcpy( self->db.db_name, self->request.header.db_name, self->request.header.db_name_len );
		self->db.db_name[self->request.header.db_name_len] = '\0';
		self->db.lock = db_lock_pool_get( self->db.db_name );
		if ( config.access_control == TRUE )
		{
			self->user.dbname = self->db.db_name;
			db_set_authorizer( &self->db, self );
		}
		else if ( repl_master_able == TRUE )
			db_set_replicate( &self->db, self );
		client_set_hook( self );

		return SQLITE_OK;
	}
	else
		log_server_print( "*** Database \"%s\" cannot open: %s", self->request.header.db_name, db_error( &(self->db) ) );
	return open_res;
}

#define R_BUF		&(self->response.buf[self->response.offset])
#define R_BUF_LIMIT	(self->response.buf_size-self->response.offset)
int client_process( struct thread_t *self, char *query )
{
	extern short	cache_able;

	int		result;
	int		change;
	long long int	rowid;
	int		rows;
	int		columns;
	char		*column_name;
	int		column_name_len;
	int		i, p;
	char		*data;
	int		data_len;

	int		header_end = 0;
	struct cache_t	qcache;

#ifdef __DEBUG
	printf("***** SQL: %s\n", query);
#endif

	self->response.start = self->response.offset = 20;

	if ( cache_able == TRUE )
	{
		qcache.sql = query;
		result = cache_get( self->db.db_name, &qcache );
		if ( result == SQLITE_OK )
		{
			//DEBUG2("cache transfering");
			self->response.offset += response_buf_open( R_BUF, R_BUF_LIMIT, 0, qcache.rowid, qcache.rows, &self->bench_start );
			memcpy( R_BUF, qcache.buf, qcache.buf_len );
			self->response.offset += qcache.buf_len;
			self->response.buf[self->response.offset] = '\0';
			self->response.status = 200;
			self->response.status_msg = "OK";
			return response( self );
		}
	}

	result = db_get_table( &self->db, query );

	if ( result != SQLITE_OK )
	{
		result = response_error( self, RESPONSE_ERR_INTERNAL, self->db.errmsg );
		db_free_table( &self->db );
		db_err_free( &self->db );
		return result;
	}

	change = db_changes( &self->db );
	rowid = db_last_rowid( &self->db );
	rows = db_table_num_rows( &self->db );

	self->response.offset += response_buf_open( R_BUF, R_BUF_LIMIT, change, rowid, rows, &self->bench_start );
	header_end = self->response.offset;

	self->response.status = 200;
	self->response.status_msg = "OK";
	//query success
	if ( db_table_has_rows( &self->db ) == SQLITE_OK )
	{
		columns = db_table_columns( &self->db );
		//columns open
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_OPEN );
		for ( i=0; i < columns; i++ )
		{
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			column_name_len = db_table_column_name( &column_name, &self->db, i );
			self->response.offset += response_xml_printf( R_BUF, column_name, column_name_len );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
		}
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_CLOSE );

		for ( i=0; i < rows; i++ )
		{
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_OPEN );
			for ( p=0; p < columns; p++ )
			{
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
				data = db_table_result( &self->db, i, p );
				if ( data != NULL )
				{
					data_len = strlen( data );
					if ( data_len + 100 > R_BUF_LIMIT )
					{
						/* chunked response start */
						result = response_flush( self, FALSE );
						if ( result != SQLITE_OK )
						{
							db_free_table( &self->db );
							return result;
						}
					}
					self->response.offset += response_xml_printf( R_BUF, data, data_len );
				}
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );
			}
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_CLOSE );
		}
	}
	else
	{
		columns = db_table_columns( &self->db );
		if ( columns > 0 )
		{
			//columns open
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_OPEN );
			for ( i=0; i < columns; i++ )
			{
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
				column_name_len = db_table_column_name( &column_name, &self->db, i );
				self->response.offset += response_xml_printf( R_BUF, column_name, column_name_len );
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
			}
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_CLOSE );
		}
	}

	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_CLOSE );
	db_free_table( &self->db );

	if ( self->response.chunked == TRUE )
		return response_flush( self, TRUE );
	else
	{
		if ( cache_able == TRUE )
		{
			if ( rows > 0 || self->db.updated == FALSE )
			{
				result = response( self );
				cache_push( self->db.db_name, query, rowid, rows, &(self->response.buf[header_end]), self->response.offset-header_end );
				return result;
			}
			else
			{
				result = response( self );
				if ( self->db.updated == TRUE )
				{
					//DEBUG2("DB updated. cache truncate")
					cache_release( self->db.db_name );
					self->db.updated = FALSE;
				}
				return result;
			}
		}
		else
			return response( self );
	}
}
#undef R_BUF
#undef R_BUF_LIMIT

void client_session_init( struct thread_t *self )
{
	self->keep_alive = FALSE;

	self->request.read_len = self->request.header_len = 0;
	self->request.data_len = /*self->request.header.node_idx = */self->request.header.method = self->request.header.db_name_len = 0;
	self->request.data = NULL;
	self->request.header.db_name = self->request.header.extend = NULL;

	self->request.header.cookie = self->request.header.auth = self->request.header.modified = self->request.header.connection = NULL;
	self->request.header.content_length = -1;

	self->response.start = self->response.offset = 0;
	self->response.status = 0;
	self->response.header_sent = self->response.set_cookie = self->response.chunked = FALSE;
	self->response.status_msg = self->response.errmsg = NULL;
}

void client_safe_exit( struct thread_t *self )
{
	if ( self->request.buf != NULL )
		free( self->request.buf );
	if ( self->response.header_buf != NULL )
		free( self->response.header_buf );
	if ( self->response.buf != NULL )
		free( self->response.buf );
	close( self->client_fd );

#ifdef	ENABLE_SSL
	if ( self->ssl != NULL )
		SSL_free( self->ssl );
#endif

	repl_master_clear( self );

	while ( self->db.pstmt_cnt > 0 )
		pstmt_internal_release( self, self->db.pstmt[0].id );

	if ( self->db.pstmt_tree != NULL )
	{
		pstmt_internal_gc( self );
		free( self->db.pstmt_tree );
	}

	if ( self->db.pstmt_buf != NULL )
		free( self->db.pstmt_buf );
	pstmt_release_all( self );

	if ( self->db.db != NULL )
	{
		if ( self->db.auth == TRUE && self->db.transaction == TRUE )
		{
			if ( self->db.auto_commit == TRUE )
				db_commit( &self->db );
			else
				db_rollback( &self->db );
		}
	}
	db_close( &self->db );

	if ( self->init == TRUE )
		monitor_thread_delete( self );
	
	free( self );
}

void client_set_hook( struct thread_t *self )
{
	extern short		cache_able;
	extern short		repl_master_able;

	if ( cache_able == TRUE || repl_master_able == TRUE )
	{
#ifdef	SQLITE_ENABLE_UPDATE_HOOK
		sqlite3_update_hook( self->db.db, client_db_updated, self );

#if defined(SQLITE_ENABLE_COMMIT_HOOK) && defined(SQLITE_ENABLE_ROLLBACK_HOOK)
		if ( repl_master_able == TRUE )
		{
			sqlite3_commit_hook( self->db.db, client_db_commited, self );
			sqlite3_rollback_hook( self->db.db, client_db_rollback, self );
		}
#endif
#endif
	}
	self->db.updated = FALSE;
}

void client_db_updated( void *data, int action, char const *db_name, char const *table_name, sqlite_int64 rowid )
{
	extern short		repl_master_able;

	if ( repl_master_able == TRUE )
		repl_master_updated( (struct thread_t *) data );
}

int client_db_commited( void *data )
{
	extern short		cache_able;
	extern short		repl_master_able;
	extern short		repl_slave_able;

	if ( cache_able == TRUE )
		cache_release_hook( &(((struct thread_t *) data)->db) );

	if ( repl_slave_able == TRUE )
		return SQLITE_ERROR;

	if ( repl_master_able == TRUE )
		return repl_master_commited( data );

	return SQLITE_OK;
}


void client_db_rollback( void *data )
{
	extern short		repl_master_able;

	if ( repl_master_able == TRUE )
		repl_master_rollback( data );
}
