#include "include/common.h"

short			repl_master_able = FALSE;
struct repl_master_t	repl_master;


int repl_master_init()
{
	FILE            *test;
	char            filename[512];

	snprintf( filename, 512, "var/replication/.init" );
	test = fopen( filename, "w" );
	if ( test == NULL )
		return SQLITE_ERROR;
	fclose( test );
	unlink( filename );

#if defined(SQLITE_ENABLE_UPDATE_HOOK) && defined(SQLITE_ENABLE_COMMIT_HOOK) && defined(SQLITE_ENABLE_ROLLBACK_HOOK)
	repl_master_able = TRUE;
#endif

	if ( repl_master_able == TRUE )
	{
		pthread_mutex_init( &repl_master.lock, NULL );

		pthread_cond_init( &repl_master.cond, NULL );
		pthread_mutex_init( &repl_master.cond_lock, NULL );

		pthread_cond_init( &repl_master.deliver_cond, NULL );
		pthread_mutex_init( &repl_master.deliver_cond_lock, NULL );

		repl_master.queue = NULL;
		repl_master.files = repl_master.files_last = NULL;
		repl_master.queue_cnt = repl_master.files_cnt = 0;

		//repl master bin writer thread
		pthread_create( &repl_master.thread, NULL, repl_master_proc, NULL );

		//repl master deliver thread
		pthread_create( &repl_master.thread, NULL, repl_master_deliver_proc, NULL );
	}
	else
		printf( "   Replication master init failed: SQLite version is lower than 3.3.5\n" );

	return SQLITE_OK;
}

#define REPL_QUEUE	repl_master.queue[i].repl_log_queue[p]
void *repl_master_proc( void *nil )
{
	extern short			terminate;
	char				*buf = NULL;
	struct repl_master_queue_t	*tmp;
	int				i, p, z;
	int				start, end;
	uint32_t			len, totallen;
	int				size;
	uint32_t			nlong;
	char				double_tmp[128];

	//4MB
	size = 1024*1024*4;
	buf = malloc( size );

	while ( 1 )
	{
		if ( repl_master.queue_cnt <= 0 )
		{
			pthread_cond_wait( &repl_master.cond, &repl_master.cond_lock );
			pthread_mutex_unlock( &repl_master.cond_lock );
		}

		start = 0;
		if ( repl_master.queue_cnt > 0 )
		{
			totallen = 0;

			pthread_mutex_lock( &repl_master.lock );
			end = repl_master.queue_cnt;
			pthread_mutex_unlock( &repl_master.lock );

repl_master_proc_loop:
			for ( i=start; i < end; i++ )
			{
				buf[0]=0x01;
				buf[1]=0x02;
				len = strlen(repl_master.queue[i].dbname);
				memcpy( &(buf[2]), repl_master.queue[i].dbname, len );
				totallen = len+3;
				buf[totallen-1]='{';
				for ( p=0; p < repl_master.queue[i].repl_log_queue_len; p++ )
				{
					buf[totallen]='[';
					totallen++;
					len = strlen( REPL_QUEUE.sql );
					nlong = htonl( len );
					if ( REPL_QUEUE.params_cnt > 0 )
					{
						//DEBUG2(REPL_QUEUE.sql);
						memcpy( &(buf[totallen]), &nlong, sizeof(nlong) );
						memcpy( &(buf[totallen+sizeof(nlong)]), REPL_QUEUE.sql, len );
						totallen += len+sizeof(nlong);
						buf[totallen]=':';
						totallen++;

						for ( z=0; z < REPL_QUEUE.params_cnt; z++ )
						{
							buf[totallen]=REPL_QUEUE.params[z].type;
							totallen++;
							nlong = htonl( REPL_QUEUE.params[z].idx );
							memcpy( &(buf[totallen]), &nlong, sizeof(nlong) );
							totallen += sizeof(nlong);
							switch ( REPL_QUEUE.params[z].type )
							{
								case 'I':
									nlong = htonl( REPL_QUEUE.params[z].data.value_integer );
									memcpy( &(buf[totallen]), &nlong, sizeof(nlong) );
									totallen += sizeof(nlong);
								break;
								case 'T':
									nlong = htonl( REPL_QUEUE.params[z].data_len );
									memcpy( &(buf[totallen]), &nlong, sizeof(nlong) );
									totallen += sizeof(nlong);
									memcpy( &(buf[totallen]),
										REPL_QUEUE.params[z].data.value_text, 
										REPL_QUEUE.params[z].data_len );
									totallen += REPL_QUEUE.params[z].data_len+1;
									buf[totallen-1]='\0';
								break;
								case 'N':
								break;
								case 'D':
									len = snprintf( double_tmp, 128, "%f",
										REPL_QUEUE.params[z].data.value_double );
									nlong = htonl( len );
									memcpy( &(buf[totallen]), &nlong, sizeof(nlong) );
									totallen += sizeof(nlong);
									memcpy( &(buf[totallen]), &double_tmp, len );
									totallen += len;
								break;
								case 'B':
									nlong = htonl( REPL_QUEUE.params[z].data_len );
									memcpy( &(buf[totallen]), &nlong, sizeof(nlong) );
									totallen += sizeof(nlong);
									memcpy( &(buf[totallen]),
										REPL_QUEUE.params[z].data.value_blob, 
										REPL_QUEUE.params[z].data_len );
									totallen += REPL_QUEUE.params[z].data_len+1;
									buf[totallen-1]='\0';
								break;
							}
						}
					}
					else
					{
						//DEBUG2(REPL_QUEUE.sql);
						memcpy( &(buf[totallen]), &nlong, sizeof(nlong) );
						memcpy( &(buf[totallen+sizeof(nlong)]), REPL_QUEUE.sql, len );
						totallen += len+sizeof(nlong);
					}
					buf[totallen]=']';
					totallen++;
				}
				buf[totallen]='}';
				buf[totallen+1]='\0';
				totallen++;

				//write to bin log file
				repl_master_write_log( buf, totallen );
			}
			pthread_mutex_lock( &repl_master.lock );

			if ( end < repl_master.queue_cnt )
			{
				start = i;
				end = repl_master.queue_cnt;
				pthread_mutex_unlock( &repl_master.lock );
				goto repl_master_proc_loop;
			}

			tmp = repl_master.queue;
			repl_master.queue = NULL;
			repl_master.queue_cnt = 0;

			pthread_mutex_unlock( &repl_master.lock );

			repl_master_flush_files();

			pthread_cond_signal( &repl_master.deliver_cond );

			while ( i > 0 && i-- )
			{
				for ( p=0; p < tmp[i].repl_log_queue_len; p++ )
					repl_master_queue_clear( &(tmp[i].repl_log_queue[p]) );
				free( tmp[i].dbname );
				free( tmp[i].repl_log_queue );
			}
			free( tmp );

		}
		if ( terminate == TRUE || repl_master_able != TRUE )
			break;
	}

	free( buf );
	pthread_exit(0);
}
#undef REPL_QUEUE

size_t repl_master_write_log( char *buf, size_t length )
{
	struct repl_master_file_t	*file;

	file = repl_master_get_file();
	if ( file == NULL )
	{
		repl_master_able = FALSE;
		return 0;
	}

	fwrite( buf, length, 1, file->file );
	file->changed = TRUE;

	return 0;
}

struct repl_master_file_t *repl_master_get_file()
{
	static int			id = 0;
	struct repl_master_file_t	*file = NULL;

	if ( repl_master.files_last != NULL && repl_master.files_last->offset < REPL_MASTER_LOG_LIMIT )
		return repl_master.files_last;
	else
	{
		file = malloc( sizeof(struct repl_master_file_t) );
		memset( file, 0, sizeof(struct repl_master_file_t) );

		file->id = ++id;
		snprintf( file->filename, 1024, "var/replication/binlog%d", file->id );

		file->file = fopen( file->filename, "w" );
		file->offset = ftell( file->file );

		if ( repl_master.files_last == NULL )
		{
			repl_master.files = repl_master.files_last = file;
			repl_master.sending_id = file->id;
			repl_master.sending_offset = file->offset;
		}
		else
		{
			repl_master.files_last->next = file;
			file->prev = repl_master.files_last;
			repl_master.files_last = file;
		}
	}

	return file;
}

void repl_master_flush_files()
{
	struct repl_master_file_t	*file;

	for ( file=repl_master.files; file; file=file->next )
	{
		if ( file->changed == TRUE )
		{
			fflush( file->file );
			file->changed = TRUE;
			file->offset = ftell( file->file );
		}
	}
}

struct repl_log_queue_t *repl_master_append( struct thread_t *self )
{
	struct repl_log_queue_t	*node = NULL;

	node = malloc( sizeof(struct repl_log_queue_t) * (self->repl_log_queue_len+1) );
	if ( node == NULL )
		return NULL;
	memset( node, 0, sizeof(struct repl_log_queue_t) * (self->repl_log_queue_len+1) );
	if ( self->repl_log_queue_len > 0 )
	{
		memcpy( node, self->repl_log_queue, sizeof(struct repl_log_queue_t) * self->repl_log_queue_len );
		free( self->repl_log_queue );
	}
	self->repl_log_queue = node;
	self->repl_log_queue_len++;

	return &(self->repl_log_queue[self->repl_log_queue_len-1]);
}

int repl_master_authorizer( void *data, int auth_code, const char *argv1, const char *argv2, const char *dbversion, const char *inner )
{
	struct thread_t *self = (struct thread_t *) data;

	switch ( auth_code )
	{
		case SQLITE_CREATE_TABLE:		/* Table Name      NULL            */
		case SQLITE_DROP_TABLE:			/* Table Name      NULL            */
		case SQLITE_CREATE_TEMP_TABLE:		/* Table Name      NULL            */
		case SQLITE_DROP_TEMP_TABLE:		/* Table Name      NULL            */
		case SQLITE_CREATE_VIEW:		/* View Name       NULL            */
		case SQLITE_DROP_VIEW:			/* View Name       NULL            */
		case SQLITE_CREATE_TEMP_VIEW:		/* View Name       NULL            */
		case SQLITE_DROP_TEMP_VIEW:		/* View Name       NULL            */
		case SQLITE_CREATE_INDEX:		/* Index Name      Table Name      */
		case SQLITE_DROP_INDEX:			/* Index Name      Table Name      */
		case SQLITE_CREATE_TEMP_INDEX:		/* Index Name      Table Name      */
		case SQLITE_DROP_TEMP_INDEX:		/* Index Name      Table Name      */
		case SQLITE_CREATE_TRIGGER:		/* Trigger Name    Table Name      */
		case SQLITE_DROP_TRIGGER:		/* Trigger Name    Table Name      */
		case SQLITE_CREATE_TEMP_TRIGGER:	/* Trigger Name    Table Name      */
		case SQLITE_DROP_TEMP_TRIGGER:		/* Trigger Name    Table Name      */
		case SQLITE_ALTER_TABLE:		/* Database Name   Table Name      */
#ifdef SQLITE_FUNCTION
		case SQLITE_FUNCTION:
#endif
			if ( self->db.work_id == self->db.work_id_curr )
				return SQLITE_OK;

			self->db.work_id_curr = self->db.work_id;

			repl_master_updated( self );
			repl_master_commited( self );
		break;
	}

	return SQLITE_OK;
}

void repl_master_updated( struct thread_t *self )
{
	struct pstmt_t		*pstmt;
	struct repl_log_queue_t	*node = NULL;

	int			i;

	if ( repl_master_able != TRUE )
		return;

	node = repl_master_append( self );
	if ( node == NULL )
		return;

	//DEBUG2("repl_master_updated");
	if ( self->request.header.method == REQUEST_METHOD_OPTION )
	{
		//prepared statement
		pstmt = &(self->db.pstmt[self->db.pstmt_run]);
		node->sql = malloc( strlen(pstmt->sql)+1 );
		strcpy( node->sql, pstmt->sql );

		node->params_cnt = pstmt->params_cnt;
		if ( node->params_cnt > 0 )
		{
			node->params = malloc( sizeof(struct pstmt_param_t) * node->params_cnt );
			memcpy( node->params, pstmt->params, sizeof(struct pstmt_param_t) * node->params_cnt );
			for ( i=0; i < node->params_cnt; i++ )
			{
				if ( pstmt->params[i].type == 'T' )
				{
					node->params[i].data.value_text = malloc( node->params[i].data_len+1 );
					memcpy( node->params[i].data.value_text, pstmt->params[i].data.value_text, node->params[i].data_len );
					node->params[i].data.value_text[node->params[i].data_len] = '\0';
				}
				else if ( pstmt->params[i].type == 'B' )
				{
					node->params[i].data.value_blob = malloc( node->params[i].data_len+1 );
					memcpy( node->params[i].data.value_blob, pstmt->params[i].data.value_blob, node->params[i].data_len );
					node->params[i].data.value_blob[node->params[i].data_len] = '\0';
				}
			}
		}
	}
	else
	{
		node->sql = malloc( strlen(self->request.data)+1 );
		strcpy( node->sql, self->request.data );
	}
}

int repl_master_commited( void *data )
{
	struct thread_t		*self = (struct thread_t *) data;
	char			*dbname;

	if ( repl_master_able != TRUE || self->repl_log_queue_len <= 0 )
		return SQLITE_OK;

	//DEBUG2("repl_master_commited");

	dbname = malloc( strlen( self->db.db_name )+1 );
	strcpy( dbname, self->db.db_name );
	pthread_mutex_lock( &repl_master.lock );

	if ( repl_master.queue == NULL )
	{
		repl_master.queue = malloc( sizeof(struct repl_master_queue_t) );
		repl_master.queue_cnt = 0;
	}
	else
		repl_master.queue = realloc( repl_master.queue, sizeof(struct repl_master_queue_t)*(repl_master.queue_cnt+1) );

	repl_master.queue[repl_master.queue_cnt].repl_log_queue_len = self->repl_log_queue_len;
	repl_master.queue[repl_master.queue_cnt].repl_log_queue = self->repl_log_queue;
	repl_master.queue[repl_master.queue_cnt].dbname = dbname;
	repl_master.queue_cnt++;

	self->repl_log_queue_len = 0;
	self->repl_log_queue = NULL;

	pthread_mutex_unlock( &repl_master.lock );

	pthread_cond_signal( &repl_master.cond );

	usleep( 1 );

	//repl_master_clear( self );

	return SQLITE_OK;
}

void repl_master_rollback( void *data )
{
	struct thread_t	*self = (struct thread_t *) data;

	//DEBUG2("repl_master_rollback");

	repl_master_clear( self );
}

void repl_master_clear( struct thread_t *self )
{
	int	i;

	if ( self->repl_log_queue_len <= 0 )
		return;

	for ( i=0; i < self->repl_log_queue_len; i++ )
		repl_master_queue_clear( &(self->repl_log_queue[i]) );

	free( self->repl_log_queue );
	self->repl_log_queue = NULL;
	self->repl_log_queue_len = 0;
}

void repl_master_queue_clear( struct repl_log_queue_t *queue )
{
	int	p;

	free( queue->sql );
	if ( queue->params_cnt > 0 )
	{
		for ( p=0; p < queue->params_cnt; p++ )
		{
			if ( queue->params[p].type == 'T' )
				free( queue->params[p].data.value_text );
			else if ( queue->params[p].type == 'B' )
				free( queue->params[p].data.value_blob );
		}
		free( queue->params );
	}
}

void repl_master_close_all_files()
{
	struct repl_master_file_t	*file;

	for ( file=repl_master.files; file; file=file->next )
	{
		fclose( file->file );
		free( file );
	}
}

void repl_master_safe_exit()
{
	int	i, p;

	if ( repl_master_able == TRUE )
	{
		for ( i=0; i < repl_master.queue_cnt; i++ )
		{
			for ( p=0; p < repl_master.queue[i].repl_log_queue_len; p++ )
				repl_master_queue_clear( &(repl_master.queue[i].repl_log_queue[p]) );
			free( repl_master.queue[i].dbname );
			free( repl_master.queue[i].repl_log_queue );
		}
		free( repl_master.queue );

		repl_master_close_all_files();

		pthread_mutex_destroy( &repl_master.lock );

		pthread_cond_destroy( &repl_master.cond );
		pthread_mutex_destroy( &repl_master.cond_lock );

		pthread_cond_destroy( &repl_master.deliver_cond );
		pthread_mutex_destroy( &repl_master.deliver_cond_lock );
	}
}

