#include "include/common.h"

void *repl_master_deliver_proc( void *nil )
{
	extern struct config_t		config;
	extern short			repl_master_able;
	extern struct repl_master_t	repl_master;
	extern short			terminate;
	extern int			errno;

	sock_t				server_addr;
	int				server_fd;
	struct pollfd			server_poll;

	sock_t				client_addr;
	socklen_t			client_addr_size;
	int				client_fd;

	struct pollfd			auth_poll;
	char				auth_str[1024];
	char				id[1024];
	char				*pass;
	struct db_user_t		db_user;

	int				client_cnt = 0;
	struct pollfd			*client_poll = NULL;

	struct timespec			condwait;

	struct repl_master_file_t	*file;
	int				fd;
	int				offset;

	int				result;
	int				i;


	server_fd = server_sock_init( &server_addr, config.replication_master_port, 1000 );
	if (server_fd <= 0)
	{
		log_server_print( "***** Replication deliver init failed: %s", strerror(errno) );
		pthread_exit(0);
	}

	server_poll.fd = server_fd;
	server_poll.events = POLLIN;

	auth_poll.events = POLLIN;

	client_addr_size = sizeof( client_addr.sa );

	client_sock_nonblock( server_fd );

	result = ETIMEDOUT;

	while ( 1 )
	{
		if ( terminate == TRUE || repl_master_able != TRUE )
			break;

		time( &condwait.tv_sec );
		condwait.tv_sec += 1;

		if ( repl_master.files_last == NULL ||
			( repl_master.sending_id == repl_master.files_last->id && repl_master.sending_offset == repl_master.files_last->offset ) )
		{
			result = pthread_cond_timedwait( &repl_master.deliver_cond, &repl_master.deliver_cond_lock, &condwait );
			pthread_mutex_unlock( &repl_master.deliver_cond_lock );
		}

		//if cond_wait timeout
		if ( result != 0 )
		{
			//new slave accept
			client_fd = accept( server_fd, &client_addr.sa, &client_addr_size );
			
			if ( client_fd > 0 )
			{
				if ( client_sock_nonblock( client_fd ) == -1 )
				{
					close( client_fd );
					continue;
				}

				auth_poll.fd = client_fd;
				if ( poll( &auth_poll, 1, 1000 ) <= 0 )
				{
					close( client_fd );
					continue;
				}

				result = read( client_fd, auth_str, 1024 );
				if ( result <= 0 )
				{
					result = 0;
					close( client_fd );
					continue;
				}
				auth_str[result] = '\0';

				result = auth_parse_str( id, &pass, auth_str );

				if ( result != SQLITE_OK )
				{
					result = 0;
					write( client_fd, "0", 1);
					close( client_fd );
					continue;
				}

				if ( auth_user_check( id, pass, &db_user ) != SQLITE_OK )
				{
					result = 0;
					write( client_fd, "0", 1);
					close( client_fd );
					continue;
				}

				if ( write( client_fd, "1", 1) <= 0 )
				{
					result = 0;
					close( client_fd );
					continue;
				}

				if ( client_poll == NULL )
					client_poll = malloc( sizeof(struct pollfd) );
				else
					client_poll = realloc( client_poll, sizeof(struct pollfd)*(client_cnt+1) );
				client_poll[client_cnt].fd = client_fd;
				client_poll[client_cnt].events = POLLOUT | POLLERR;
				client_cnt++;
			}
			result = 0;
			continue;
		}

		if ( repl_master.files_last == NULL ||
			( repl_master.sending_id == repl_master.files_last->id && repl_master.sending_offset == repl_master.files_last->offset ) )
			continue;

		if ( client_cnt <= 0 )
		{
			repl_master.sending_id = repl_master.files_last->id;
			repl_master.sending_offset = repl_master.files_last->offset;
			continue;
		}

		if ( repl_master.sending_id == repl_master.files_last->id )
		{
			fd = open( repl_master.files_last->filename, O_RDONLY );
			if ( fd < 0 )
				continue;

			for ( i=0; i < client_cnt; i++ )
			{
				if ( client_poll[i].fd < 0 )
					continue;

				//printf("%zd, %zd\n", repl_master.sending_offset, repl_master.files_last->offset );

				if ( net_sendfile( client_poll[i].fd, fd, repl_master.sending_offset, repl_master.files_last->offset - repl_master.sending_offset ) != SQLITE_OK )
				{
					close( client_poll[i].fd );
					client_poll[i].fd = -1;
				}
			}
			repl_master.sending_offset = repl_master.files_last->offset;

			close( fd );
		}
		else
		{
			file = repl_master_get_file_by_id( repl_master.sending_id );
			if ( file == NULL )
				continue;

			for ( ; file; file=file->next )
			{
				fd = open( file->filename, O_RDONLY );
				if ( fd < 0 )
					continue;

				if ( file->id == repl_master.sending_id )
					offset = repl_master.sending_offset;
				else
					offset = 0;

				for ( i=0; i < client_cnt; i++ )
				{
					if ( client_poll[i].fd < 0 )
						continue;
					if ( net_sendfile( client_poll[i].fd, fd, offset, file->offset - offset ) != SQLITE_OK )
					{
						close( client_poll[i].fd );
						client_poll[i].fd = -1;
					}
				}
				close( fd );

				repl_master.sending_id = file->id;
				repl_master.sending_offset = file->offset;
			}
		}
	}

	close( server_fd );

	for ( i=0; i < client_cnt; i++ )
	{
		if ( client_poll[i].fd < 0 )
		continue;
		close( client_poll[i].fd );
	}
	if ( client_cnt > 0 )
		free( client_poll );

	pthread_exit(0);
}

struct repl_master_file_t *repl_master_get_file_by_id( int id )
{
	extern struct repl_master_t	repl_master;
	struct repl_master_file_t	*file = NULL;

	if ( repl_master.files == NULL || repl_master.files_last == NULL )
		return NULL;
	else
	{
		for ( file=repl_master.files_last; file; file=file->prev )
		{
			if ( file->id == id )
				return file;
		}
	}

	return NULL;
}
