#include "include/common.h"

short			repl_slave_able = FALSE;
short			repl_slave_thread_init = FALSE;
pthread_t		repl_slave_thread;

void repl_slave_init()
{
#if defined(SQLITE_ENABLE_COMMIT_HOOK)
	repl_slave_able = TRUE;
#endif

	if ( repl_slave_able == TRUE )
	{
		pthread_create( &repl_slave_thread, NULL, repl_slave_proc, NULL );
	}
	else
		printf( "   Replication slave init failed: SQLite version is lower than 3.3.5\n" );
}

void *repl_slave_proc( void *nil )
{
	extern struct config_t	config;
	extern int		errno;

	struct hostent		*hostent;
	int			sockfd;
	struct sockaddr_in	addr;
	struct pollfd		pollfd;
	int			poll_res;

	char			auth_str[1024];

	char			*buffer;
	char			*buf;
	int			length;
	int			res;

	if ( repl_slave_able != TRUE )
		pthread_exit(0);

	//connect to master
	if ( (hostent = gethostbyname( config.replication_slave_host )) == NULL )
	{
		perror("* Cannot connect to replication master");
		log_server_print( "***** Cannot connect to replication master: %s", strerror(errno) );
		goto slave_end;
	}

	if ( ( sockfd = socket( AF_INET, SOCK_STREAM, 0 ) ) == -1 )
	{
		perror("* Cannot connect to replication master");
		log_server_print( "***** Cannot connect to replication master: %s", strerror(errno) );
		goto slave_end;
	}

	addr.sin_family = AF_INET;
	addr.sin_port = htons( config.replication_slave_port );
	addr.sin_addr = *((struct in_addr *)hostent->h_addr);

	if ( config.replication_slave_user == NULL || strlen(config.replication_slave_user) <= 0
		|| config.replication_slave_pass == NULL || strlen(config.replication_slave_pass) <= 0
		|| _sqlited_make_auth_str( auth_str, config.replication_slave_user, config.replication_slave_pass ) != SQLITE_OK )
	{
		close( sockfd );
		printf("* Cannot connect to replication master: cannot make auth info\n");
		log_server("* Cannot connect to replication master: cannot make auth info");
		goto slave_end;
	}

	if ( connect( sockfd, (struct sockaddr *)&addr, sizeof(struct sockaddr) ) == -1 )
	{
		close( sockfd );
		perror("* Cannot connect to replication master");
		log_server_print( "***** Cannot connect to replication master: %s", strerror(errno) );
		goto slave_end;
        }

	if ( write( sockfd, auth_str, strlen(auth_str) ) <= 0 )
	{
		close( sockfd );
		perror("* Cannot connect to replication master");
		log_server_print( "***** Cannot connect to replication master: %s", strerror(errno) );
		goto slave_end;
	}

	if ( read( sockfd, auth_str, 1024 ) <= 0 )
	{
		close( sockfd );
		perror("* Cannot connect to replication master");
		log_server_print( "***** Cannot connect to replication master: %s", strerror(errno) );
		goto slave_end;
	}

	if ( auth_str[0] != '1' )
	{
		close( sockfd );
		printf("* Replication master authorization Failed.\n");
		log_server("* Replication master authorization Failed.");
		goto slave_end;
	}

	pollfd.fd = sockfd;
	pollfd.events = POLLIN | POLLERR | POLLHUP;

	repl_slave_thread_init = TRUE;

	buffer = malloc( REPL_SLAVE_BUF );
	buf = buffer;

	length = 0;
	while ( 1 )
	{
		poll_res = poll( &pollfd, 1, 1000 );

		if ( poll_res == 0 )
			continue;
		else if ( poll_res < 0 )
		{
			free( buf );
			break;
		}

		res = read( sockfd, buf, REPL_SLAVE_BUF );
		if ( res <= 0 )
			break;

		length += res;

		if ( buf[res-1] != '}' )
		{
			buf += res;
			continue;
		}

		buf = buffer;

		if ( repl_slave_apply( buffer, length ) != SQLITE_OK )
		{
			length = 0;
			printf("* Replication failed.\n");
			log_server("* Replication failed.");
			continue;
		}

		length = 0;
	}

	free( buffer );
	close( sockfd );

slave_end:
	repl_slave_thread_init = TRUE;

	if ( repl_slave_able == TRUE )
	{
		sleep( 1 );
		pthread_create( &repl_slave_thread, NULL, repl_slave_proc, NULL );
	}

	pthread_exit(0);
}

#define LOG_PARAM	log.queue[i].params[p]
int repl_slave_apply( char *buf, int length )
{
	struct repl_slave_db_t		log;
	struct db_t			db;
	int				res;
	int				offset = 0;
	int				i, p;

	memset( &log, 0, sizeof(struct repl_slave_db_t) );
	memset( &db, 0, sizeof(struct db_t) );

	while ( length > offset )
	{
		if ( repl_slave_get_record( &log, buf, &res ) != SQLITE_OK )
			return SQLITE_ERROR;

		buf += res;
		offset += res;

		if ( log.queue_len > 0 )
		{
			if ( db_open_init( &db, log.dbname, NULL ) == SQLITE_OK )
			{
				for ( i=0; i < log.queue_len; i++ )
				{
					if ( log.queue[i].params_cnt > 0 )
					{
						if ( db_statement( &db, log.queue[i].sql ) == SQLITE_OK )
						{
							for ( p=0; p < log.queue[i].params_cnt; p++ )
							{
								switch ( LOG_PARAM.type )
								{
									case 'I':
										if ( sqlite3_bind_int( db.stmt, LOG_PARAM.idx, LOG_PARAM.data.value_integer ) != SQLITE_OK )
										{
											p += 999999;
											db_err_free( &db );
										}
									break;
									case 'T':
										if ( sqlite3_bind_text( db.stmt, LOG_PARAM.idx, LOG_PARAM.data.value_text, LOG_PARAM.data_len, SQLITE_TRANSIENT ) != SQLITE_OK )
										{
											p += 999999;
											db_err_free( &db );
										}
									break;
									case 'N':
										if ( sqlite3_bind_null( db.stmt, LOG_PARAM.idx ) != SQLITE_OK )
										{
											p += 999999;
											db_err_free( &db );
										}
									break;
									case 'D':
										if ( sqlite3_bind_double( db.stmt, LOG_PARAM.idx, LOG_PARAM.data.value_double ) != SQLITE_OK )
										{
											p += 999999;
											db_err_free( &db );
										}
									break;
									case 'B':
										if ( sqlite3_bind_blob( db.stmt, LOG_PARAM.idx, LOG_PARAM.data.value_blob, LOG_PARAM.data_len, SQLITE_TRANSIENT ) != SQLITE_OK )
										{
											p += 999999;
											db_err_free( &db );
										}
									break;
								}
							}

							while ( sqlite3_step( db.stmt ) == SQLITE_ROW ) {}
							db_statement_free( &db );
							db_err_free( &db );
						}
					}
					else
					{
						db_get_table( &db, log.queue[i].sql );
						db_free_table( &db );
						db_err_free( &db );
					}

					if ( log.queue[i].params != NULL )
						free( log.queue[i].params );
				}
				// cache reset
				cache_release( log.dbname );

				db_close( &db );
			}

			free( log.queue );
			log.queue = NULL;
			log.queue_len = 0;
		}
	}

	return SQLITE_OK;
}
#undef LOG_PARAM

int repl_slave_get_record( struct repl_slave_db_t *db, char *buf, int *offset )
{
	char			*tmp;
	uint32_t		size;
	struct repl_log_queue_t	*queue;
	int			parsed = 0;
	short			is_pstmt;

	if ( buf[0] != 0x01 || buf[1] != 0x02 )
		return SQLITE_ERROR;

	parsed += 2;

	db->dbname = &(buf[2]);
	tmp = strchr( db->dbname, '{' );

	if ( tmp == NULL )
		return SQLITE_ERROR;

	tmp[0] = '\0';
	db->dbname_len = strlen( db->dbname );
	buf = ++tmp;

	parsed += db->dbname_len+1;

	while ( 1 )
	{

		if ( buf[0] != '[' )
			return SQLITE_ERROR;

		if ( db->queue == NULL )
		{
			db->queue = malloc( sizeof(struct repl_log_queue_t) );
			db->queue_len = 0;
			queue = db->queue;
		}
		else
		{
			db->queue = realloc( db->queue, sizeof(struct repl_log_queue_t)*(db->queue_len+1) );
			queue = &(db->queue[db->queue_len]);
		}

		db->queue_len++;
		buf++;
		parsed++;

		memset( queue, 0, sizeof(struct repl_log_queue_t) );

		memcpy( &size, buf, sizeof(uint32_t) );
		size = ntohl( size );

		buf += sizeof(uint32_t);
		parsed += sizeof(uint32_t);

		queue->sql = buf;

		buf += size;
		parsed += size;

		if ( buf[0] == ':' )
			is_pstmt = TRUE;
		else
			is_pstmt = FALSE;

		buf[0] = '\0';
		buf++;
		parsed++;

		//DEBUG2(queue->sql);

		if ( is_pstmt == TRUE )
		{
			while ( buf[0] != '}' ) 
			{
				if ( buf[0] != 'I' && buf[0] != 'T' && buf[0] != 'N' && buf[0] != 'D' && buf[0] != 'B' )
				{
					buf++;
					parsed++;
					continue;
				}

				if ( queue->params == NULL )
				{
					queue->params = malloc( sizeof(struct pstmt_param_t) );
					queue->params_cnt = 0;
				}
				else
					queue->params = realloc( queue->params, sizeof(struct pstmt_param_t)*(queue->params_cnt+1) );

				queue->params[queue->params_cnt].type = buf[0];
				buf++;
				parsed++;

				memcpy( &queue->params[queue->params_cnt].idx, buf, sizeof(uint32_t) );
				queue->params[queue->params_cnt].idx = (int) ntohl( queue->params[queue->params_cnt].idx );
				buf += sizeof(uint32_t);
				parsed += sizeof(uint32_t);

				switch ( queue->params[queue->params_cnt].type )
				{
					case 'I':
						memcpy( &queue->params[queue->params_cnt].data.value_integer, buf, sizeof(uint32_t) );
						queue->params[queue->params_cnt].data.value_integer = ntohl( queue->params[queue->params_cnt].data.value_integer );
						buf += sizeof(uint32_t);
						parsed += sizeof(uint32_t);
					break;
					case 'T':
						memcpy( &size, buf, sizeof(uint32_t) );
						size = ntohl( size );

						buf += sizeof(uint32_t);
						queue->params[queue->params_cnt].data_len = size;

						queue->params[queue->params_cnt].data.value_text = buf;
						buf += size+1;
						parsed += size+1+sizeof(uint32_t);
					break;
					case 'N':
					break;
					case 'D':
						memcpy( &size, buf, sizeof(uint32_t) );
						size = ntohl( size );

						buf += sizeof(uint32_t);
						queue->params[queue->params_cnt].data_len = size;

						queue->params[queue->params_cnt].data.value_double = strtod( buf, NULL );
						buf += size;
						parsed += size+sizeof(uint32_t);
					break;
					case 'B':
						memcpy( &size, buf, sizeof(uint32_t) );
						size = ntohl( size );

						buf += sizeof(uint32_t);
						queue->params[queue->params_cnt].data_len = size;

						queue->params[queue->params_cnt].data.value_blob = buf;
						buf += size+1;
						parsed += size+1+sizeof(uint32_t);
					break;
				}

				queue->params_cnt++;
			}
		}

		if ( buf[0] == '}' )
		{
			parsed++;
			break;
		}
	}

	(*offset) = parsed;

	return SQLITE_OK;
}
