#include "include/common.h"

int request_header_parse( struct request_t *request )
{
	int		header_len;

	char		*req;
	size_t		idx;

	request->header_len = 0;
	for ( header_len=1; header_len < request->read_len; header_len++ )
	{
		if ( ( request->buf[header_len-1] == '\n' && request->buf[header_len] == '\n' )
			|| ( header_len > 2 && request->buf[header_len-3] == '\r' && request->buf[header_len-2] == '\n' && request->buf[header_len-1] == '\r' && request->buf[header_len] == '\n' ) )
		{
			header_len++;
			request->header_len = header_len;
			break;
		}
	}

	request->buf[request->header_len-1] = '\0';

	if ( request->header_len <= 0 )
		return SQLITE_OK;

	switch (request->buf[0])
	{
		case 'p':
		case 'P':
			if ( strncasecmp( &(request->buf[1]), "OST ", 4 ) == 0 )
			{
				request->header.method = REQUEST_METHOD_POST;
				req = &(request->buf[5]);
			}
			else if ( strncasecmp( &(request->buf[1]), "UT ", 3 ) == 0 )
			{
				request->header.method = REQUEST_METHOD_PUT;
				req = &(request->buf[4]);
			}
			else
				return SQLITE_ERROR;
		break;
		case 'g':
		case 'G':
			if ( strncasecmp( &(request->buf[1]), "ET ", 3 ) == 0 )
			{
				request->header.method = REQUEST_METHOD_GET;
				req = &(request->buf[4]);
			}
			else
				return SQLITE_ERROR;
		break;
		case 'h':
		case 'H':
			if ( strncasecmp( &(request->buf[1]), "EAD ", 4 ) == 0 )
			{
				request->header.method = REQUEST_METHOD_HEAD;
				req = &(request->buf[5]);
			}
			else
				return SQLITE_ERROR;
		break;
		case 'o':
		case 'O':
			if ( strncasecmp( &(request->buf[1]), "PTION ", 6 ) == 0 )
			{
				request->header.method = REQUEST_METHOD_OPTION;
				req = &(request->buf[7]);
			}
			else
				return SQLITE_ERROR;
		break;
		case 'd':
		case 'D':
			if ( strncasecmp( &(request->buf[1]), "ELETE ", 6 ) == 0 )
			{
				request->header.method = REQUEST_METHOD_DELETE;
				req = &(request->buf[7]);
			}
			else
				return SQLITE_ERROR;
		break;
		default:
			return SQLITE_ERROR;
		break;
	}

	if ( req == NULL )
		return SQLITE_ERROR;
	idx = strcspn( req, " " );
	if ( idx < 0 )
		return SQLITE_ERROR;
	req++;

	request->header.db_name = req;
	request->header.db_name_len = idx-1;
	request->header.db_name[request->header.db_name_len] = '\0';
	//printf("DB: (%s)\n", request->header.db_name);

	request->header.query = strchr( request->header.db_name, '?' );
	if ( request->header.query != NULL )
	{
		request->header.query[0] = '\0';
		request->header.query++;
	}

	req += request->header.db_name_len + 9;

	if ( request->header.method == REQUEST_METHOD_GET )
	{
		// connection, modified
		while ( req[0] != '\0' && req[1] != '\0' )
		{
			req += strcspn( req, "\r\n" ) + 1;
			if ( req[0] == '\r' || req[0] == '\n' )
				req++;

			if ( ( req[0] == 'c' || req[0] == 'C' ) && ( req[1] == 'o' || req[1] == 'O' ) )
			{
				if ( strncasecmp( &(req[2]), "nnection", 8 ) == 0 )
					request->header.connection = &(req[10+strspn( &(req[10]), " :" )]);
			}
			else if ( ( req[0] == 'i' || req[0] == 'I' ) && ( req[1] == 'f' || req[1] == 'F' ) )
			{
				if ( strncasecmp( &(req[2]), "-Modified-Since", 15 ) == 0 )
					request->header.modified = &(req[17+strspn( &(req[17]), " :" )]);
			}
		}

		if ( request->header.connection != NULL )
			request_rtrim( request->header.connection );

		if ( request->header.modified != NULL )
			request_rtrim( request->header.modified );
	}
	else
	{
		// content_length, cookie, connection, auth
		while ( req[0] != '\0' && req[1] != '\0' )
		{
			req += strcspn( req, "\r\n" ) + 1;
			if ( req[0] == '\r' || req[0] == '\n' )
				req++;

			if ( ( req[0] == 'c' || req[0] == 'C' ) && ( req[1] == 'o' || req[1] == 'O' ) )
			{
				if ( strncasecmp( &(req[2]), "ntent-length", 12 ) == 0 )
					request->header.content_length = atoi( &(req[14+strspn( &(req[14]), " :" )]) );
				else if ( strncasecmp( &(req[2]), "okie", 4 ) == 0 )
					request->header.cookie = &(req[6+strspn( &(req[6]), " :" )]);
				else if ( strncasecmp( &(req[2]), "nnection", 8 ) == 0 )
					request->header.connection = &(req[10+strspn( &(req[10]), " :" )]);
			}
			else if ( ( req[0] == 'a' || req[0] == 'A' ) && ( req[1] == 'u' || req[1] == 'U' ) )
			{
				if ( strncasecmp( &(req[2]), "thorization", 11 ) == 0 )
					request->header.auth = &(req[13+strspn( &(req[13]), " :" )]);
			}
		}

		if ( request->header.cookie != NULL )
			request_rtrim( request->header.cookie );

		if ( request->header.connection != NULL )
			request_rtrim( request->header.connection );

		if ( request->header.auth != NULL )
			request_rtrim( request->header.auth );
	}

	return SQLITE_OK;
}

void request_rtrim( char *buf )
{
	char	*tmp;
	tmp = strchr( buf, '\r' );
	if ( tmp != NULL )
		tmp[0] = '\0';
	tmp = strchr( buf, '\n' );
	if ( tmp != NULL )
		tmp[0] = '\0';
}

int request_end( struct request_t *request )
{
	int	content_length;

	content_length = request->header.content_length;
	if ( content_length >= 0 )
	{
		if ( request->read_len >= content_length + request->header_len )
		{
			request->data = &(request->buf[request->header_len]);
			if ( memcmp( request->data, "query=", 6) == 0 )
			{
				request->data += 6;
				content_length -= 6;
			}

			if ( request->header_len > 0 )
				content_length = request_decode( request->data, request->data, content_length );
			//printf("%s\n", request->data);
			request->data_len = content_length;
			return SQLITE_OK;
		}
		return SQLITE_EMPTY;
	}
	return SQLITE_ERROR;
}

int request_decode( char *to, char *from, int from_len )
{
	int	len;

	for ( len=0; from_len--; ++to, ++from, ++len )
	{
		if ( from[0] == '%' && from[1] == '%')
		{
			--to;
			--len;
			continue;
		}

		if ( from[0] == '%' && isxdigit( from[1] ) && isxdigit( from[2] ) )
		{
			*to = request_decode_hexit( from[1] ) * 16 + request_decode_hexit( from[2] );
			from += 2;
			from_len -= 2;
		}
		else if ( from[0] == '+' )
			*to = ' ';
		else
			*to = *from;
	}
	*to = '\0';
	return len;
}

int request_decode_hexit( char c )
{
	if ( c >= '0' && c <= '9' )
		return c - '0';
	if ( c >= 'a' && c <= 'f' )
		return c - 'a' + 10;
	if ( c >= 'A' && c <= 'F' )
		return c - 'A' + 10;
	return 0;
}

short request_keep_alive( struct request_t *request )
{
	if ( request->header.connection != NULL )
		return ( strncasecmp( request->header.connection, "KEEP-ALIVE", 10 ) == 0 )?TRUE:FALSE;

	return FALSE;
}

