#include "include/common.h"

static short	response_errstatus[] = {
	400, 401, 500, 406, 406, 406, 406, 406, 406, 406, 404, 406, 406, 406, 406, 406
};

static char	*response_errmsg[] = {
	"Bad Request",			"Bad Request",
	"Unauthorized",			"Authorization Required",
	"Internal Server Error",	"Internal Server Error",
	"Cannot create database",	"Not Acceptable",
	"Database not found",		"Not Acceptable",
	"Cannot drop database",		"Not Acceptable",
	"Cannot create user",		"Not Acceptable",
	"User not found",		"Not Acceptable",
	"Cannot drop user",		"Not Acceptable",
	"Cannot alter user",		"Not Acceptable",
	"Resource not found",		"Not Found",
	"Cannot grant",			"Not Acceptable",
	"Cannot revoke",		"Not Acceptable",
	"File already exists",		"Not Acceptable",
	"Cannot PUT file",		"Not Acceptable",
	"Cannot DELETE file",		"Not Acceptable"
};

static char	*response_infomsg[] = {
	"Database created",
	"Database droped",
	"User created",
	"User droped",
	"User alterd",
	"Login Success",
	"GRANT execute success",
	"REVOKE execute success"
};

int response_error( struct thread_t *self, int status, char *errmsg )
{
	self->response.status = response_errstatus[status];
	self->response.status_msg = response_errmsg[(status*2)+1];
	if ( errmsg == NULL )
		self->response.errmsg = response_errmsg[status*2];
	else
		self->response.errmsg = errmsg;

	self->response.start = 0;
	self->response.offset = sprintf( self->response.buf, XML_ERROR_OPEN );
	self->response.offset += response_xml_printf( &(self->response.buf[self->response.offset]), self->response.errmsg, strlen(self->response.errmsg) );
	self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_ERROR_CLOSE );

	//printf("%s\n", self->response.buf);

	if ( self->response.chunked == TRUE )
		return response_flush( self, TRUE );
	else
		return response( self );
}

int response_info( struct thread_t *self, int status, char *infomsg )
{
	self->response.status = 202;
	self->response.status_msg = "Accepted";
	if ( infomsg == NULL )
		self->response.errmsg = response_infomsg[status];
	else
		self->response.errmsg = infomsg;

	self->response.start = 0;
	self->response.offset = sprintf( self->response.buf, XML_INFO_OPEN );
	self->response.offset += response_xml_printf( &(self->response.buf[self->response.offset]), self->response.errmsg, strlen(self->response.errmsg) );
	self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_INFO_CLOSE );

	//printf("%s\n", self->response.buf);

	if ( self->response.chunked == TRUE )
		return response_flush( self, TRUE );
	else
		return response( self );
}

int response_message( struct thread_t *self, char *msg )
{
	self->response.status = 200;
	self->response.status_msg = "OK";
	self->response.errmsg = msg;

	self->response.start = 0;
	self->response.offset = sprintf( self->response.buf, XML_INFO_OPEN );
	self->response.offset += response_xml_printf( &(self->response.buf[self->response.offset]), self->response.errmsg, strlen(self->response.errmsg) );
	self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_INFO_CLOSE );

	//printf("%s\n", self->response.buf);

	if ( self->response.chunked == TRUE )
		return response_flush( self, TRUE );
	else
		return response( self );
}

int response( struct thread_t *self )
{
	extern struct config_t	config;
	int			len;
	char			date_822[48];
	int			ret;

	if ( self->response.header_sent != TRUE )
	{
		if ( self->response.chunked == TRUE )
		{
			http_time2asctime( date_822, self->response.stat.st_mtime );
			len = snprintf( self->response.header_buf, config.buf_size_header, "HTTP/1.1 %d %s\nServer: %s\nLast-Modified: %s\nContent-type: text/xml\nTransfer-Encoding: Chunked\n", self->response.status, self->response.status_msg, SERVER_SOFTWARE, date_822 );
		}
		else if ( self->response.status == 304 )
		{
			http_time2asctime( date_822, self->response.stat.st_mtime );
			len = snprintf( self->response.header_buf, config.buf_size_header, "HTTP/1.1 %d %s\nServer: %s\nLast-Modified: %s\n", self->response.status, self->response.status_msg, SERVER_SOFTWARE, date_822 );
		}
		else if ( self->response.file > 0 )
		{
			http_time2asctime( date_822, self->response.stat.st_mtime );
#if defined(__APPLE__)
			len = snprintf( self->response.header_buf, config.buf_size_header, "HTTP/1.1 %d %s\nServer: %s\nContent-length: %lld\nLast-Modified: %s\n", self->response.status, self->response.status_msg, SERVER_SOFTWARE, self->response.stat.st_size, date_822 );
#else
			len = snprintf( self->response.header_buf, config.buf_size_header, "HTTP/1.1 %d %s\nServer: %s\nContent-length: %ld\nLast-Modified: %s\n", self->response.status, self->response.status_msg, SERVER_SOFTWARE, self->response.stat.st_size, date_822 );
#endif
		}
		else
		{
			len = snprintf( self->response.header_buf, config.buf_size_header, "HTTP/1.1 %d %s\nServer: %s\nContent-type: text/xml\nContent-length: %d\n", self->response.status, self->response.status_msg, SERVER_SOFTWARE, self->response.offset-self->response.start );
		}

		if ( self->response.status == 401 )
			len += snprintf( &(self->response.header_buf[len]), config.buf_size_header-len, "WWW-Authenticate: Basic realm=SQLiteDBMS\n" );

		if ( self->response.set_cookie )
			len += snprintf( &(self->response.header_buf[len]), config.buf_size_header-len, "Set-Cookie: SQLSESSION=%s; expire=0; path=/\n", self->session_key );

		if ( self->keep_alive == TRUE )
			len += snprintf( &(self->response.header_buf[len]), config.buf_size_header-len, "Connection: Keep-Alive\nKeep-Alive: timeout=%d\n\n", config.http_keeptime );
		else
			len += snprintf( &(self->response.header_buf[len]), config.buf_size_header-len, "Connection: Close\n\n" );

#ifdef	ENABLE_SSL
		if ( config.ssl == TRUE )
		{
			if ( net_send_ssl( self->client_fd, self->ssl, self->response.header_buf, len ) != SQLITE_OK )
				return SQLITE_ERROR_SEND;
		}
		else
		{
#endif
			if ( net_send( self->client_fd, self->response.header_buf, len ) != SQLITE_OK )
				return SQLITE_ERROR_SEND;
#ifdef	ENABLE_SSL
		}
#endif

		self->response.header_sent = TRUE;

#ifdef __DEBUG
		//printf("%s", self->response.header_buf);
#endif
	}

	if ( self->request.header.method == REQUEST_METHOD_HEAD || self->response.status == 304 )
	{
		log_http( self->client_ip, self->request.header.method, self->user.id, self->request.header.db_name, self->response.status, 0 );
		return SQLITE_OK;
	}

	if ( self->response.file > 0 )
	{
		lseek( self->response.file, 0, self->response.file_offset );
#ifdef	ENABLE_SSL
		if ( config.ssl == TRUE )
			ret = net_sendfile_ssl( self->client_fd, self->ssl, self->response.file, self->response.file_offset, self->response.stat.st_size );
		else
#endif
			ret = net_sendfile( self->client_fd, self->response.file, self->response.file_offset, self->response.stat.st_size );
	}
	else
	{
#ifdef __DEBUG
		//printf("%s\n", &(self->response.buf[self->response.start]));
#endif

#ifdef	ENABLE_SSL
		if ( config.ssl == TRUE )
			ret = net_send_ssl( self->client_fd, self->ssl, &(self->response.buf[self->response.start]), (self->response.offset-self->response.start) );
		else
#endif		
			ret = net_send( self->client_fd, &(self->response.buf[self->response.start]), (self->response.offset-self->response.start) );
	}

	if ( self->request.header.method == REQUEST_METHOD_POST )
	{
		if ( config.log_sql == TRUE )
			log_db( self->client_ip, self->request.header.method, self->user.id, self->request.header.db_name, self->request.data, self->response.status, (self->response.offset-self->response.start) );
	}
	else if ( self->request.header.method == REQUEST_METHOD_OPTION )
	{
		if ( config.log_sql == TRUE && self->db.pstmt_run > 0 && self->db.pstmt[self->db.pstmt_run].nolog == FALSE )
			log_db( self->client_ip, self->request.header.method, self->user.id, self->request.header.db_name, self->db.pstmt[self->db.pstmt_run].sql, self->response.status, (self->response.offset-self->response.start) );
	}
	else
		log_http( self->client_ip, self->request.header.method, self->user.id, self->request.header.db_name, self->response.status, self->response.stat.st_size );
	return ret;

}

int response_buf_open( char *buf, size_t size, int changes, long long int rowid, int rows, struct timeval *start )
{
	struct timeval	end;
	long		sec;
	long		usec;

	gettimeofday( &end, NULL );
	if ( end.tv_usec < start->tv_usec )
	{
		end.tv_usec += 1000000;
		end.tv_sec++;
	}
	sec = end.tv_sec - start->tv_sec;
	usec = end.tv_usec - start->tv_usec;

	return snprintf( buf, size, XML_RESPONSE_OPEN, changes, rowid, rows, sec, usec );
}

int response_flush( struct thread_t *self, short is_last )
{
	int	result;
	char	chunk[16];
	int	chunk_len;

	if ( self->response.chunked != TRUE )
		self->response.chunked = TRUE;

	chunk_len = snprintf( chunk, 16, "%x", (self->response.offset-self->response.start) );

	sprintf( &(self->response.buf[(self->response.start - (chunk_len + 2))]), "%s\r", chunk );
	(self->response.buf[self->response.start-1]) = '\n';
	self->response.start -= (chunk_len + 2);

	(self->response.buf[self->response.offset]) = '\r';
	(self->response.buf[self->response.offset+1]) = '\n';
	self->response.offset += 2;

	if ( is_last == TRUE )
	{
		(self->response.buf[self->response.offset]) = '0';
		(self->response.buf[self->response.offset+1]) = '\r';
		(self->response.buf[self->response.offset+2]) = '\n';
		(self->response.buf[self->response.offset+3]) = '\r';
		(self->response.buf[self->response.offset+4]) = '\n';
		(self->response.buf[self->response.offset+5]) = '\0';
		self->response.offset += 5;
	}

	result = response( self );
	if ( result == SQLITE_OK )
		self->response.start = self->response.offset = 20;

	return result;
}

int response_xml_printf( char *dst, const char *src, const int src_len )
{
	int		offset;
	int		i;

	offset = i = 0;
	for ( ; i < src_len; i++ )
	{
		switch ( src[i] )
		{
			case '&':
				dst[offset++] = '&';
				dst[offset++] = 'a';
				dst[offset++] = 'm';
				dst[offset++] = 'p';
				dst[offset++] = ';';
			break;
			case '"':
				dst[offset++] = '&';
				dst[offset++] = 'q';
				dst[offset++] = 'u';
				dst[offset++] = 'o';
				dst[offset++] = 't';
				dst[offset++] = ';';
			break;
			case '\'':
				dst[offset++] = '&';
				dst[offset++] = '#';
				dst[offset++] = '0';
				dst[offset++] = '3';
				dst[offset++] = '9';
				dst[offset++] = ';';
			break;
			case '<':
				dst[offset++] = '&';
				dst[offset++] = 'l';
				dst[offset++] = 't';
				dst[offset++] = ';';
			break;
			case '>':
				dst[offset++] = '&';
				dst[offset++] = 'g';
				dst[offset++] = 't';
				dst[offset++] = ';';
			break;
			default:
				dst[offset++] = src[i];
			break;
		}
	}
	return offset;
}
