#include "include/common.h"

#define decode(c) if(c >= 'A' && c <= 'Z') c  = c - 'A'; \
		else if(c >= 'a' && c <= 'z') c  = c - 'a' + 26; \
		else if(c >= '0' && c <= '9') c  = c - '0' + 52; \
		else if(c == '+')             c  = 62; \
		else if(c == '/')             c  = 63; \
		else                          c  = 0; \

int base64_decode( char *dst, const char *src, unsigned int len )
{
	int x, y = 0;
	char triple[3];
	char quad[4];

	len = strlen(src);

	for( x = 0; x < len; x += 4 )
	{
		memset(quad, 0, 4);
		memcpy(quad, &src[x], 4 - (len - x) % 4);

		decode(quad[0]);
		decode(quad[1]);
		decode(quad[2]);
		decode(quad[3]);

		triple[0] = (quad[0] << 2) | quad[1] >> 4;
		triple[1] = ((quad[1] << 4) & 0xF0) | quad[2] >> 2;
		triple[2] = ((quad[2] << 6) & 0xC0) | quad[3];

		memcpy(&dst[y], triple, 3);            
		y += 3;      
	}
	return y;
}
#undef decode

static char base64_table[] = {
	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
	'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/', '\0'
};

int _sqlited_make_auth_str( char *auth_str, char *username, char *passwd )
{
	char	*tmp;
	char	*src;
	size_t	length;
	int	i = 0;

	length = strlen( username ) + 2 + strlen( passwd );
	tmp = (char *) malloc( length );
	if ( tmp == NULL )
		return SQLITE_NOMEM;

	snprintf( tmp, length, "%s:%s", username, passwd );
	src = tmp;

	if ( auth_str == NULL )
	{
		free( tmp );
		return SQLITE_ERROR;
	}

	length--;

	while (length > 2) {
		auth_str[i++] = base64_table[src[0] >> 2];
		auth_str[i++] = base64_table[((src[0] & 0x03) << 4) + (src[1] >> 4)];
		auth_str[i++] = base64_table[((src[1] & 0x0f) << 2) + (src[2] >> 6)];
		auth_str[i++] = base64_table[src[2] & 0x3f];

		src += 3;
		length -= 3;
	}

	switch ( length )
	{
		case 2:
			auth_str[i++] = base64_table[src[0] >> 2];
			auth_str[i++] = base64_table[((src[0] & 0x03) << 4) + (src[1] >> 4)];
			auth_str[i++] = base64_table[(src[1] & 0x0f) << 2];
			auth_str[i++] = '=';
		break;
		case 1:
			auth_str[i++] = base64_table[src[0] >> 2];
			auth_str[i++] = base64_table[(src[0] & 0x03) << 4];
			auth_str[i++] = '=';
			auth_str[i++] = '=';
		break;
	}

	auth_str[i] = '\0';
	free( tmp );

	return SQLITE_OK;
}

